package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** currently this class exports only static methods to create repetitions; there are no public instance methods or constructors */
public class Repeat extends Union {

    /** repeat zero or one times */
    public  static Repeat maybe(Element e)              { return new Repeat(e, true, false); }
    /** repeat zero or more times */
    public  static Repeat many0(Element e)              { return new Repeat(e, true, true); }
    /** repeat zero or more times, separated by <tt>sep</tt> */
    public  static Repeat many0(Element e, Element sep) { return new Repeat(e, true, true, sep); }
    /** repeat one or more times */
    public  static Repeat many1(Element e)              { return new Repeat(e, false, true); }
    /** repeat one or more times, separated by <tt>sep</tt> */
    public  static Repeat many1(Element e, Element sep) { return new Repeat(e, false, true, sep); }

    /** repeat zero or more times, matching a maximal sequence of atoms */
    public  static Repeat maximal0(Element e)              { return new Repeat(e, true, true, null, true); }
    /** repeat one or more times, matching a maximal sequence of atoms */
    public  static Repeat maximal1(Element e)              { return new Repeat(e, false, true, null, true); }
    /** repeat one or more times, separated by an atom <tt>sep</tt>, matching a maximal sequence */
    public  static Repeat maximal1(Element e, Element sep) { return new Repeat(e, false, true, sep, true); }

    private Repeat(final Element e, boolean zeroOkay, boolean manyOkay) { this(e, zeroOkay, manyOkay, null); }
    private Repeat(final Element e, boolean zeroOkay, boolean manyOkay, Element separator) { this(e, zeroOkay, manyOkay, separator, false); }
    private Repeat(final Element e, boolean zeroOkay, boolean manyOkay, final Element separator, boolean maximal) {
        super(e+(!manyOkay ? "?" : (zeroOkay ? (maximal ? "**" : "*") : (maximal ? "++" : "+")))+(separator==null?"":("/"+separator)), true);
        if (maximal && zeroOkay && separator != null)
            throw new RuntimeException("cannot create a maximal repetition of zero or more items with a separator (yet): " + this);
        if (zeroOkay) {
            add(new Sequence.Constant.Empty());
            if (manyOkay) add(new Sequence.Singleton(many1(e, separator)));
            else          add(new Sequence.Singleton(e));
        } else {
            add(new Sequence.RewritingSequence(null, new Element[] { e }));
            if (separator==null)
                add(new Sequence.Unwrap(new Element[] { e,                 Repeat.this }));
            else
                add(new Sequence.Unwrap(new Element[] { e, separator,      Repeat.this }, new boolean[] { false, true, false }));
        }
        if (maximal) for(Sequence s : this) s.noFollow = separator==null ? e : separator;
    }
}
